/*
 * Copyright 2008, Systemic Realtime Design LLC. All Rights Reserved.
 *
 * This source code may contain confidential information of Systemic
 * Realtime Design LLC and its licensors. Any use, reproduction,
 * modification, disclosure, distribution or transfer of this software,
 * or any software which includes or is based upon any of this code, is
 * prohibited unless expressly authorized by SRD by written agreement. For
 * more information (including whether this source code file has been
 * published) please email licensing@sysrtime.com.
 */

/*
 * net/tipc/tipc_eth_media.c: Ethernet bearer support for TIPC
 * 
 * Copyright (c) 2001-2006, Ericsson AB
 * Copyright (c) 2005-2007, Wind River Systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the names of the copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */


#include <sys/netmgr.h>
#include <sys/socket.h>
#include <net/if_dl.h>
#include <net/ethertypes.h>
#include <net/if_types.h>
#include <net/if_ether.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <sys/io-net.h>
#include <errno.h>
#include "tipc_core.h"
#include "sys/param_bsd.h"
#include "sys/mbuf.h"
#include "tipc_dbg.h"

#include <net/tipc/tipc.h>
#include <net/tipc/tipc_plugin_if.h>
#include <net/tipc/tipc_plugin_msg.h>

#define MAX_ETH_BEARERS		TIPC_MAX_BEARERS
#define ETH_LINK_PRIORITY	TIPC_DEF_LINK_PRI
#define ETH_LINK_TOLERANCE	TIPC_DEF_LINK_TOL

#ifndef ETH_ALEN
#define ETH_ALEN 6
#endif
#define ETH_HLEN 14

#ifndef ETH_P_TIPC
#define ETH_P_TIPC 0x88CA
#endif

/* QNX io-pkt ethernet interface */

typedef struct {
	uint8_t    dest[6];
	uint8_t    source[6];
	uint16_t   type;
	uint8_t    payload[0];   // total 1500 bytes, 1496 avail to user
} ether_packet_t;


extern int etype_list_insert(unsigned etype,
                             struct ifqueue *(*input_func)(struct ifnet *ifp, struct mbuf *m));
extern int etype_list_remove(unsigned etype);

static struct ifnet* iopkt_find_l4_en (const char *if_name);


/**
 * struct eth_bearer - Ethernet bearer data structure
 * @bearer: ptr to associated "generic" bearer structure
 * @dev: ptr to associated Ethernet network device
 * @tipc_packet_type: used in binding TIPC to Ethernet driver
 * @cleanup: work item used when disabling bearer
 */
 
struct eth_bearer {
	struct tipc_bearer *bearer;
	struct ifnet *ifp;
};

static struct tipc_media eth_media_info;
static struct eth_bearer eth_bearers[MAX_ETH_BEARERS];
static int eth_started = 0;


/*
 * Declare public routines so "sparse" doesn't complain ...
 */

extern int tipc_eth_media_start(void);
extern void tipc_eth_media_stop(void);

/**
 * eth_media_addr_init - initialize Ethernet media address structure
 * 
 * Structure's "value" field stores address info in the following format:
 * - Ethernet media type identifier [4 bytes, in network byte order]
 * - MAC address [6 bytes]
 * - unused [10 bytes of zeroes]
 * 
 * Note: This is the same format as the TIPC neighbour discovery message uses
 * to designate an Ethernet address, which simplies the job of getting the
 * media address into/out of the message header.
 */

static void eth_media_addr_init(struct tipc_media_addr *a, char *mac)
{
	memset(a->value, 0, sizeof(a->value));
	a->value[3] = TIPC_MEDIA_ID_ETH;
	memcpy(&a->value[4], mac, ETH_ALEN);
	
	a->media_id = TIPC_MEDIA_ID_ETH;
	a->broadcast = !memcmp(mac, &eth_media_info.bcast_addr.value[4], ETH_ALEN);
}

	
/**
 * send_msg - send a TIPC message out over an Ethernet interface 
 */

static int send_msg(struct sk_buff *buf, struct tipc_bearer *tb_ptr, 
		    struct tipc_media_addr *dest)
{
	int status;
	struct eth_bearer *eb_ptr;
	struct sockaddr eth_raddr;
	ether_packet_t *eth_pkt;
	mbuf_t *mbuf_dup;

	eb_ptr = (struct eth_bearer *)(tb_ptr->usr_handle);

	/*
	 * Create duplicate of message for sending, since TIPC needs to keep
	 * the original for possible retransmission
	 * (note that only a single mBlk is duplicated, not the entire message)
	 */

	mbuf_dup = MBUF_DUP (buf->mBufPtr, 0, M_COPYALL, M_DONTWAIT);
	if (mbuf_dup == NULL)
	{
		printk("Can't duplicate buffer to send over Ethernet! (errno=0x%08x)\n", errno);

		/* Indicate success & let link handle message loss */
		return EOK;
	}

	/* prepare ether_raddr */
	eth_raddr.sa_family = AF_UNSPEC;
	eth_raddr.sa_len    = sizeof(eth_raddr);
	eth_pkt = (ether_packet_t *)&eth_raddr.sa_data;
	memcpy(eth_pkt->dest, (unchar*)&dest->value[4], 6);
	eth_pkt->type = htons(ETH_P_TIPC);

#if 0
	/* leave space so ether_output() could fill ether header in */
	mbuf_dup->m_data += sizeof(ether_packet_t);
	mbuf_dup->m_len  -= sizeof(ether_packet_t);
	mbuf_dup->m_pkthdr.len -= sizeof(ether_packet_t);
#endif

	/* pass mbuf down to driver for tx (driver frees the mbuf) */
	status = eb_ptr->ifp->if_output (eb_ptr->ifp, mbuf_dup, &(eth_raddr), NULL);
	if (status != TX_DOWN_OK)
	{
		if (status == TX_DOWN_AGAIN)
			printk("%s: network driver is busy\n", __FUNCTION__);
		else if (status == TX_DOWN_FAILED)
			printk("%s: network driver transmit failed\n", __FUNCTION__);
		else	
			printk("%s: unknown network driver error %d\n", __FUNCTION__, status);
	}
	return status;
}



/**
 * recv_msg - handle incoming TIPC message from an Ethernet interface
 * 
 * Accept only packets explicitly sent to this node, or broadcast packets;
 * ignores packets sent using Ethernet multicast, and traffic sent to other
 * nodes (which can happen if interface is running in promiscuous mode).
 */

static struct ifqueue * recv_msg (struct ifnet *ifp, mbuf_t *mbuf)
{
	struct sk_buff *buf;                /* message buffer */
	ether_packet_t *eth_pkt;
	int eth_pkt_len;	
	struct eth_bearer *eb_ptr;          /* Ethernet bearer used */
	int i;

	/* this should be a MH_PKTHDR */
	if (!(mbuf->m_flags & M_PKTHDR)) {
		MBUF_FREE(mbuf);
		return NULL;
	}
	
	/* find our eth_bearer */
	for (i=0; i<MAX_ETH_BEARERS; i++)
	{
		eb_ptr = (struct eth_bearer*) &eth_bearers[i];
		if (eb_ptr && (eb_ptr->ifp == ifp))
			break;
	}

	if ((i>=MAX_ETH_BEARERS) || (eb_ptr->bearer == NULL)) {
		err("%s: eth_bearer not found!\n", __FUNCTION__);
	    MBUF_FREE(mbuf);
	    return NULL;
	}
	
	/* range check eth_pkt_len for runt packets */
	eth_pkt_len = mbuf->m_pkthdr.len;
	if (eth_pkt_len < sizeof(ether_packet_t))
	{
		err("bad rxd packet - too short: %d\n", eth_pkt_len);
		MBUF_FREE(mbuf);
		return NULL;
	}

	/* point to raw ethernet packet with 2 mac addrs and 2 type bytes */
	eth_pkt = mtod(mbuf, ether_packet_t *);

	/* check if type is TIPC */
	if (eth_pkt->type != htons(ETH_P_TIPC))
	{
		err("bad rx packet ethernet type: %x\n", eth_pkt->type);
		MBUF_FREE(mbuf);
		return NULL;	
	}		

	/*
	 * Accept only packets explicitly sent to this node, or broadcast packets;
	 * ignore packets sent using Ethernet multicast, and traffic sent to other
	 * nodes (which can happen if interface is running in promiscuous mode)
	 */
	if ((memcmp(eth_pkt->dest, &eb_ptr->bearer->addr.value[4], ETH_ALEN)) && 
		(memcmp(eth_pkt->dest, "\xff\xff\xff\xff\xff\xff", ETH_ALEN)))
	{
		MBUF_FREE(mbuf);
		return NULL;
	}

	/* move data ptr after ethernet header */
	mbuf->m_data = eth_pkt->payload;
	mbuf->m_len -= sizeof(ether_packet_t);
	mbuf->m_pkthdr.len -= sizeof(ether_packet_t);

	/* Allocate message buffer to hold incoming message */

	buf = vxskb_attach (mbuf);
	if (buf == NULL)
	{
		printk ("Out of buffers; incoming Ethernet message discarded\n");
		MBUF_FREE(mbuf);
		return NULL;
	}

	/* Hand off message buffer for processing */

	tipc_recv_msg (buf, eb_ptr->bearer);

	return NULL;
}


/**
 * enable_bearer - attach TIPC bearer to an Ethernet interface 
 */

static int enable_bearer (struct tipc_bearer *tb_ptr)
{
	struct eth_bearer *eb_ptr = &eth_bearers[0];
	struct eth_bearer *stop = &eth_bearers[MAX_ETH_BEARERS];
	struct ifnet *ifp;
	char   *if_name;
	char   bearer_name[TIPC_MAX_BEARER_NAME];
	int status;

	/* Get the name of interface to use */ 

	memcpy(bearer_name, tb_ptr->name, TIPC_MAX_BEARER_NAME);
	if_name = strchr(bearer_name, ':');
	if (if_name == NULL) {
		err("%s: invalid bearer name format: %s\n", __FUNCTION__, tb_ptr->name);
		return -ENODEV;
	}
	*(if_name++) = 0;

	
	/* Find the ifnet ethernet structure associated with this driver */
	 
	ifp = iopkt_find_l4_en (if_name);
	if (ifp == NULL) {
		err("%s: can't find interface %s\n", __FUNCTION__, if_name);
		return -ENODEV;
	}

	/* Create Ethernet bearer for device */

	while (eb_ptr->ifp != NULL) {
		if (++eb_ptr == stop)
			return -EDQUOT;
	}

	/* Associate TIPC bearer with Ethernet bearer */

	eb_ptr->bearer = tb_ptr;
	eb_ptr->ifp = ifp;
	tb_ptr->usr_handle = (void *)eb_ptr;
	tb_ptr->mtu = ifp->if_mtu;
	tb_ptr->blocked = 0;

	/* setup mac address */
	eth_media_addr_init(&tb_ptr->addr, 
						(char *)(ifp->if_sadl->sdl_data + ifp->if_sadl->sdl_nlen));

	/* Bind TIPC input routine to i/f & populate fields of Ethernet bearer entry */
	status = etype_list_insert (ETH_P_TIPC, recv_msg);
	if (status != 0) {
		err("%s: etype_list_insert: %s\n", __FUNCTION__, strerror(status));
	}
	return TIPC_OK;
}




/**
 * disable_bearer - detach TIPC bearer from an Ethernet interface 
 *
 * Mark Ethernet bearer as inactive so that incoming buffers are thrown away,
 * then get worker thread to complete bearer cleanup.  (Can't do cleanup
 * here because cleanup code needs to sleep and caller holds spinlocks.)
 */

static void disable_bearer(struct tipc_bearer *tb_ptr)
{
	struct eth_bearer *eb_ptr = (struct eth_bearer *)tb_ptr->usr_handle;
	if (eb_ptr->ifp == NULL)
		return;

	etype_list_remove(ETH_P_TIPC);
	memset(eb_ptr, 0, sizeof(struct eth_bearer));
}

#if 0 // TODO: is there an equivalent notification scheme from QNX io-pkt?
/**
 * recv_notification - handle device updates from OS
 *
 * Change the state of the Ethernet bearer (if any) associated with the 
 * specified device.
 */

static int recv_notification(struct notifier_block *nb, unsigned long evt, 
			     void *dv)
{
	struct net_device *dev = (struct net_device *)dv;
	struct eth_bearer *eb_ptr = &eth_bearers[0];
	struct eth_bearer *stop = &eth_bearers[MAX_ETH_BEARERS];

	while ((eb_ptr->dev != dev)) {
		if (++eb_ptr == stop)
			return NOTIFY_DONE;	/* couldn't find device */
	}
	if (!eb_ptr->bearer)
		return NOTIFY_DONE;		/* bearer had been disabled */

        eb_ptr->bearer->mtu = dev->mtu;

	switch (evt) {
	case NETDEV_CHANGE:
		if (netif_carrier_ok(dev))
			tipc_continue(eb_ptr->bearer);
		else
			tipc_block_bearer(eb_ptr->bearer->name);
		break;
	case NETDEV_UP:
		tipc_continue(eb_ptr->bearer);
		break;
	case NETDEV_DOWN:
		tipc_block_bearer(eb_ptr->bearer->name);
		break;
	case NETDEV_CHANGEMTU:
        case NETDEV_CHANGEADDR:
		tipc_block_bearer(eb_ptr->bearer->name);
                tipc_continue(eb_ptr->bearer);
		break;
	case NETDEV_UNREGISTER:
        case NETDEV_CHANGENAME:
		tipc_disable_bearer(eb_ptr->bearer->name);
		break;
	}
	return NOTIFY_OK;
}
#endif

static int eth_msg2addr(struct tipc_media_addr *a, u32 *msg_area)
{
	if (msg_area[0] != htonl(TIPC_MEDIA_ID_ETH))
		return 1;

	eth_media_addr_init(a, (char *)&msg_area[1]);
	return 0;
}

static int eth_addr2msg(struct tipc_media_addr *a, u32 *msg_area)
{
	if (a->media_id != TIPC_MEDIA_ID_ETH)
		return 1;

	memcpy(msg_area, a->value, sizeof(a->value));
	return 0;
}

/**
 * eth_addr2str - convert Ethernet address to string
 */

static int eth_addr2str(struct tipc_media_addr *a, char *str_buf, int str_size)
{                       
	unsigned char *mac;

	if ((a->media_id != TIPC_MEDIA_ID_ETH) || (str_size < 18))
		return 1;
	
	mac = (unsigned char *)&a->value[4];
	sprintf(str_buf, "%02x:%02x:%02x:%02x:%02x:%02x",
		mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
	return 0;
}


/**
 * eth_str2addr - convert string to Ethernet address
 */

static int eth_str2addr(struct tipc_media_addr *a, char *str_buf)
{                     
	char mac[6];

	if (ETH_ALEN != sscanf(str_buf, "%02x:%02x:%02x:%02x:%02x:%02x",
				(u32 *)&mac[0], (u32 *)&mac[1], (u32 *)&mac[2],
				(u32 *)&mac[3], (u32 *)&mac[4], (u32 *)&mac[5]))
	return 1;

	eth_media_addr_init(a, mac);
	return 0;
}

/*
 * Ethernet media registration info required by TIPC
 */

static struct tipc_media eth_media_info = {
	TIPC_MEDIA_ID_ETH,
	"eth",
	ETH_LINK_PRIORITY,
	ETH_LINK_TOLERANCE,
	TIPC_DEF_LINK_WIN,
	send_msg,
	enable_bearer,
	disable_bearer,
	eth_addr2str,
	eth_str2addr,
        eth_msg2addr,
        eth_addr2msg,
	{{0, 0, 0, TIPC_MEDIA_ID_ETH, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	TIPC_MEDIA_ID_ETH, 1}
};


/**
 * tipc_eth_media_start - activate Ethernet bearer support
 *
 * Register Ethernet media type with TIPC bearer code.
 * Also register with OS for notifications about device state changes.
 */

int tipc_eth_media_start(void)
{                       
	int res;

	if (eth_started)
		return -EINVAL;

	memset(eth_bearers, 0, sizeof(eth_bearers));

	res = tipc_register_media(&eth_media_info);
	if (res)
		return res;

	eth_started = 1;
	return res;
}

/**
 * tipc_eth_media_stop - deactivate Ethernet bearer support
 */

void tipc_eth_media_stop(void)
{
	struct ifnet *ifp;
	
	if (!eth_started)
		return;

#if 0 /* TODO: not needed? */
	/* Find the ifnet ethernet structure associated with this driver 
	 *
	 * TODO: use driver_name to find specific ethernet driver instead of 
	 *       just grabbing the first ethernet interface? 
	 */
	ifp = iopkt_find_l4_en();
	if (ifp == NULL) {
		return;
	}
	
	/* take it down */
	if (ifp->if_drain)
		ifp->if_drain(ifp);
	if (ifp->if_stop)
		ifp->if_stop(ifp, 1);
#endif
	
	eth_started = 0;
}


/*******************************************************************************
*
* iopkt_find_l4_en - find ethernet interface
*
*/

static struct ifnet* iopkt_find_l4_en (const char *if_name)
{
	extern struct ifnet_head ifnet;
	struct ifnet *ifp = NULL;

	/* Populate the dp(s) now, any driver mounted later, is not used. */
	info("Looking for interface %s\n", if_name);	
	for (ifp = TAILQ_FIRST(&ifnet); ifp; ifp = TAILQ_NEXT(ifp, if_list))
	{
		/* skip non IFT_ETHER interface */
		if (!ifp->if_sadl || ifp->if_sadl->sdl_type != IFT_ETHER)
		{
			continue;
		}

		/* is this user specified interface we want? */
		if (strcmp(ifp->if_xname, if_name) != 0) {
			info("Ignoring interface %s\n", ifp->if_xname);	
			continue;
		}
			
		/* let's make sure the interface is running */
		if ((ifp->if_flags & (IFF_RUNNING | IFF_UP)) != (IFF_RUNNING | IFF_UP)) {
			int error;
			error = (*ifp->if_init)(ifp);
			if (error != 0) {
				warn("%s(): can't init interface '%s': %p",
							__FUNCTION__, ifp->if_xname, ifp);
				return NULL;
			}
			ifp->if_flags |= IFF_UP;
		}
		return ifp;
	}
	return NULL;
}

