/*
 * Copyright 2008, Systemic Realtime Design LLC. All Rights Reserved.
 *
 * This source code may contain confidential information of Systemic
 * Realtime Design LLC and its licensors. Any use, reproduction,
 * modification, disclosure, distribution or transfer of this software,
 * or any software which includes or is based upon any of this code, is
 * prohibited unless expressly authorized by SRD by written agreement. For
 * more information (including whether this source code file has been
 * published) please email licensing@sysrtime.com.
 */

/*****************************************************************************
* Tabs 5 9
* The carets in the following lines will match up if tabs are set properly
*   ^   ^   ^   ^   ^   ^   <== tabs
*   ^   ^   ^   ^   ^   ^   <== spaces
******************************************************************************
*
*   Copyright (C) 2007 Systemic Realtime Design, LLC.    All Rights Reserved.
*
*   This source code may contain confidential information of
*            Systemic Realtime Design, LLC.   (SRD)
*   Any use, reproduction, modification, disclosure, distribution or
*   transfer of this software, or any software which includes or is based
*   upon any of this code, is prohibited unless expressly authorized
*   by SRD by written agreement.
*
*   For more information please email licensing@SysRTime.com.
*
*****************************************************************************/

#include <stdio.h>
#include <inttypes.h>
#include <time.h>
#include <pthread.h>
#include <sys/syspage.h>
#include <sys/neutrino.h>
#include <sys/kernel.h>
#include "tipc_core.h"
#include "tipc_timer.h"
#include "tipc_dbg.h"
#include "receive.h"



/****************************************************************************
*
* k_init_timer - initialize a timer
*
* This routine initializes a timer structure with the arguments provided.
* Timer must be initialized before use (and terminated when no longer needed).
*
* Parameters:
*   timer    - pointer to timer structure
*   routine  - pointer to routine to invoke when timer expires
*   argument - value to pass to routine when timer expires
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* SEE ALSO: N/A
*
* \NOMANUAL
*
*/

void k_init_timer (struct timer_list *timer, Handler routine, unsigned long argument)
{
	/* init timer struct */
	callout_init (&timer->timerRef);

	/* setup func and args */
   	callout_setfunc (&timer->timerRef, (void*)routine, (void*)argument); 
	//printf("%s: init timer %x\n", __FUNCTION__, (int)&timer->timerRef);
}



/****************************************************************************
*
* k_cancel_timer - cancel a timer
*
* Prevents a previously initialized timer from timing out.
* Can be safely called if timer is already inactive.
*
* Parameters:
*   timer    - pointer to timer structure
*
* WARNING: Calling this routine does not guarantee that timeout routine
* is not currently executing!
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* SEE ALSO: N/A
*
* \NOMANUAL
*
*/

void k_cancel_timer(struct timer_list *timer)
{
	callout_stop(&timer->timerRef);	
}



/****************************************************************************
*
* k_start_timer - start a timer
*
* Schedules a previously initialized timer for later execution.
* If timer is already running, the new timeout overrides the previous request.
*
* To ensure the timer does not expire until after the specified delay elapses,
* the amount of delay is rounded up when converting to ticks and
* then an additional tick is added to account for the fact that
* the starting time may be in the middle of the current tick.
*
* Parameters:
*   timer    - pointer to timer structure
*   msec     - time to delay (ms)
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* SEE ALSO: N/A
*
* \NOMANUAL
*
*/

void k_start_timer(struct timer_list *timer, unsigned long msec)
{
	int ticks;
	extern int hz; 

	/* check for deactivatin value */
	if (msec == 0xffffffff)
	{
		callout_stop(&timer->timerRef);	
  		return;	
	} 
   
	/* setup timer */
	ticks = msec * hz / 1000 + 1;
	
	/* callout_schedule can't handle higher than an integer */
	if (ticks < 0)
		ticks = 0x7fffffff;

	//printf("%s: timer %x expires in %d msec (%d ticks)\n", __FUNCTION__, (int)&timer->timerRef, msec, ticks);
	callout_schedule (&timer->timerRef, ticks);
}




/****************************************************************************
*
* k_term_timer - terminate a timer
*
* Prevents further use of a previously initialized timer.
*
* Parameters:
*   timer    - pointer to timer structure
*
* WARNING: Caller must ensure that the timer is not currently running
*
* (Do not "enhance" this routine to automatically cancel an active timer,
* otherwise deadlock can arise when a timeout routine calls k_term_timer.)
*
* RETURNS: N/A
*
* ERRNO: N/A
*
* SEE ALSO: N/A
*
* \NOMANUAL
*
*/

void k_term_timer (struct timer_list *timer)
{
    dbg("terminating timer %p\n", timer);
	callout_stop(&timer->timerRef);	
}




/*
 *  Timer utility functions
 */
unsigned int jiffies_to_msecs (const unsigned long j)
{
	uint32_t handler_clk_rate = SYSPAGE_ENTRY(qtime)->cycles_per_sec;

	if (handler_clk_rate < 1000)
	{
		return ((1000 / handler_clk_rate) * j);
	}
		else if (j > handler_clk_rate)
	{
		return ((j / handler_clk_rate) * 1000);
	}
	else
	{
		return (j * 1000 / handler_clk_rate);
	}
}

